@@ -26,16 +26,17 @@ Only single output problems are handled.
 # License: BSD 3 clause
 
 
-import numbers
-
+import warnings
 import pandas as pd
 from joblib import Parallel, delayed
 from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
 from sklearn.base import is_classifier, is_regressor
 from sklearn.utils.multiclass import check_classification_targets
 from sklearn.utils.validation import check_is_fitted  # , check_random_state
+from tqdm import trange
 
 from randomsubgroups.pysubgrouputils import *
+from randomsubgroups.subgroup import SubgroupPredictor
 
 
 # This function '_get_n_samples_bootstrap' is taken from sklearn.ensemble._forest
@@ -85,44 +86,46 @@ class SubgroupPredictorBase(BaseEstimator):
     """
 
     def __init__(self,
-                 n_estimators=100,
-                 # *,
-                 max_depth=1,
+                 n_estimators=200,
+                 max_depth=2,
                  max_features="auto",
                  bootstrap=True,
-                 # oob_score=False,
-                 n_jobs=None,
-                 # random_state=None,
-                 verbose=0,
-                 # class_weight=None,
                  max_samples=None,
-                 search_strategy='bestfirst',
-                 result_set_size=1,
-                 intervals_only=True,
-                 n_bins=5):
+                 n_jobs=None,
+                 verbose=1,
+                 search_strategy='static',
+                 top_n=1,
+                 result_set_size=5,
+                 intervals_only=False,
+                 n_bins=5,
+                 binning='ef'):
 
+        # Ensemble parameters
         self.n_estimators = n_estimators
         self.max_depth = max_depth
-        self.bootstrap = bootstrap
-        #         self.oob_score = oob_score
         self.max_features = max_features
-        self.n_jobs = n_jobs
-        # self.random_state = random_state
-        self.verbose = verbose
-        #         self.class_weight = class_weight
+        self.bootstrap = bootstrap
         self.max_samples = max_samples
+
+        # General parameters
+        self.n_jobs = n_jobs
+        self.verbose = verbose
+
+        # Subgroup Discovery parameters
         self.search_strategy = search_strategy
+        self.top_n = top_n
         self.result_set_size = result_set_size
         self.intervals_only = intervals_only
         self.n_bins = n_bins
+        self.binning = binning
 
         self.is_fitted_ = False
         self.estimators_ = None
         self.classes_ = None
         self.n_classes_ = None
-        self.target = None
         self.n_features_ = None
         self.n_samples = None
+        self.default_prediction = None
 
         self.column_names = None
 
@@ -142,7 +145,7 @@ class SubgroupPredictorBase(BaseEstimator):
                                  "Allowed string values are 'auto', "
                                  "'sqrt' or 'log2'.")
         elif self.max_features is None:
-            max_features = self.n_features_
+            max_features = None
         elif isinstance(self.max_features, numbers.Integral):
             max_features = self.max_features
         else:  # float
@@ -157,17 +160,18 @@ class SubgroupPredictorBase(BaseEstimator):
     def _check_estimator_output(self, result):
 
         if len(result.to_descriptions()) > 0:
-            # Use the top ``n`` subgroup in the list result.to_descriptions()
-            if self.result_set_size > 1:
-                # n = np.random.choice(len(result.to_descriptions()))
+
+            if len(result.to_descriptions()) > self.top_n:
+                # Randomly select from the the top_n subgroups in the list result.to_descriptions()
+                n = np.random.choice(self.top_n, 1, replace=False)[0]
+                # return [result.to_descriptions()[i] for i in n]
+            elif self.result_set_size > 1:
                 n = np.random.choice(len(result.to_descriptions()))
             else:
                 n = 0
-            return result.to_descriptions()[n]
+            return [result.to_descriptions()[n]]
         else:
-            # TODO: Allow that some estimators cannot find subgroups
-            msg = "Could not find subgroups for one or more estimators."
-            raise ValueError(msg)
+            return None
 
     def _apply(self, xy):
         """
@@ -194,29 +198,29 @@ class SubgroupPredictorBase(BaseEstimator):
             )
             xy = xy.sample(n=n_samples_bootstrap, replace=True)
 
-        if self.search_strategy == 'bestfirst':
+        if self.search_strategy in ['static', 'dynamic']:
+            self.ps_algorithm = LightBestFirstSearch(max_features=self.max_features, n_bins=self.n_bins,
+                                                     intervals_only=self.intervals_only,
+                                                     specialization=self.search_strategy,
+                                                     binning=self.binning)
+            subgroup_predictor = self._subgroup_discovery(xy)
+            return subgroup_predictor
+        elif self.search_strategy == 'bestfirst':
             self.ps_algorithm = ps.BestFirstSearch()
-        elif self.search_strategy == 'bestfirst2':
-            self.ps_algorithm = ps.BestFirstSearch2()
-        elif self.search_strategy == 'bestfirst3':
-            self.ps_algorithm = ps.BestFirstSearch3(self.max_features)
         elif self.search_strategy == 'beam':
             self.ps_algorithm = ps.BeamSearch()
         elif self.search_strategy == 'apriori':
             self.ps_algorithm = ps.Apriori()
         else:
             msg = "Unknown search strategy. Available options are: " \
-                  "['bestfirst', 'bestfirst2', 'bestfirst3', 'beam', 'apriori']"
+                  "['static', 'dynamic', 'beam', 'apriori']"
             raise ValueError(msg)
 
-        if self.search_strategy == 'bestfirst3':
-            subgroup, target = self._subgroup_discovery(xy)
-        else:
-            subset_columns = np.append(np.random.choice(self.n_features_, size=self.max_features, replace=False),
-                                       self.n_features_)
-            subgroup, target = self._subgroup_discovery(xy.iloc[:, subset_columns].copy())
+        subset_columns = np.append(np.random.choice(self.n_features_, size=self.max_features, replace=False),
+                                   self.n_features_)
+        subgroup_predictor = self._subgroup_discovery(xy.iloc[:, subset_columns].copy())
 
-        return subgroup, target
+        return subgroup_predictor
 
     def fit(self, x, y):
         """
@@ -237,11 +241,9 @@ class SubgroupPredictorBase(BaseEstimator):
         """
         # random_state = check_random_state(self.random_state)
 
-        is_classification = is_classifier(self)
-
-        if is_classification:
+        if is_classifier(self):
             check_classification_targets(y)
-            y = np.copy(y)
+            # y = np.copy(y)
 
             # Store the classes seen during fit and encode y
             self.classes_, y_encoded = np.unique(y, return_inverse=True)
@@ -249,12 +251,16 @@ class SubgroupPredictorBase(BaseEstimator):
             self.n_classes_ = len(self.classes_)
 
             y = y_encoded
+        elif is_regressor(self):
+            self.default_prediction = np.mean(y)
 
         # Check that x is a dataframe, if not then
         # creates a dataframe to be used in the Subgroup Discovery task
         if not isinstance(x, pd.DataFrame):
-            self.column_names = ['Col' + str(i) for i in range(1, x.shape[1] + 1)]
+            self.column_names = ['Col' + str(i) for i in range(0, x.shape[1])]
             x = pd.DataFrame.from_records(x, columns=self.column_names)
+        else:
+            self.column_names = x.columns
 
         xy = x.copy()
         xy['target'] = y
@@ -262,13 +268,16 @@ class SubgroupPredictorBase(BaseEstimator):
         self.n_samples, self.n_features_ = x.shape
         self.get_max_n_features()
 
-        model_desc = Parallel(n_jobs=self.n_jobs, verbose=self.verbose, backend='loky')(
-            delayed(self._apply)(xy.copy()) for _ in range(self.n_estimators))
+        model_desc = Parallel(n_jobs=self.n_jobs, verbose=0, backend='loky')(
+            delayed(self._apply)(xy.copy()) for _ in (trange(self.n_estimators)
+                                                      if self.verbose else range(self.n_estimators)))
 
-        self.target = np.array([target for dsc, target in model_desc])
-        self.estimators_ = [dsc for dsc, target in model_desc]
+        self.estimators_ = [SubgroupPredictor.from_dict(sg) for subgroup in model_desc
+                            if subgroup is not None for sg in subgroup]
 
-        self.estimators_ = [encode_subgroup(sg) for sg in self.estimators_]
+        n = self.n_estimators - len(self.estimators_)
+        if n > 0 and self.verbose:
+            print("Could not find {} out of {} estimators.".format(n, self.n_estimators))
 
         self.is_fitted_ = True
 
@@ -279,50 +288,56 @@ class SubgroupPredictorBase(BaseEstimator):
         # Check if fit had been called
         check_is_fitted(self)
 
-        is_classification = is_classifier(self)
-        if is_classification:
-            targets = self.target
+        if is_classifier(self):
+            [print(f"Target: {self.classes_[estimator.target]}; Model: {estimator}") for estimator in
+             sorted(self.estimators_, key=lambda e: e.target)]
         else:
-            targets = self.target[:, 0]
-
-        id_sorted_targets = [x for x, y in sorted(enumerate(targets), key=lambda x: x[1])]
-
-        for i in id_sorted_targets:
-            target = round(targets[i], 3)
-            model = self.estimators_[i]
-            print(f"Target: {target}; Model: {model}")
+            [print(f"Target: {estimator.target}; Model: {estimator}") for estimator in
+             sorted(self.estimators_, key=lambda e: e.target)]
 
     def show_decision(self, x):
 
         # Check if fit had been called
         check_is_fitted(self)
 
-        x = pd.DataFrame(x).transpose()
-        covered = []
-        target_distribution = []
+        # Input validation
+        # x = pd.DataFrame(x).transpose()
+        if not isinstance(x, pd.DataFrame):
+            # Create a dataframe to be used in the Subgroup Discovery task
+            x = pd.DataFrame.from_records((x,), columns=self.column_names)
 
-        is_classification = is_classifier(self)
-        if is_classification:
-            target = self.target
-        else:
-            target = self.target[:, 0]
+        covered_estimators = filter(lambda e: e.covers(x), self.estimators_)
+        covered_estimators = list(sorted(covered_estimators, key=lambda e: (e.target, e)))
 
-        for i, estimator in enumerate(self.estimators_):
-            if estimator.covers(x):
-                covered.append(i)
-                target_distribution.append(target[i])
+        if len(covered_estimators) > 0:
 
-        if len(covered) > 0:
-            # pd.Series(covered).hist()
-            print("The predicted value is:", self.predict(x))
-            print("From a total of", len(covered), "estimators.\n")
+            print("The predicted value is:", self.predict(x)[0])
+            print("From a total of", len(covered_estimators), "estimators.\n")
 
             print("The subgroups used in the prediction are:")
-            for i in covered:
-                print(self.estimators_[i], "--->", target[i])
 
-            print("\nThe targets of the subgroups used in the prediction have the following distribution:")
-            pd.Series(target_distribution).hist()
+            if is_classifier(self):
+                target_distribution = []
+                initial_target = -1
+                for estimator in covered_estimators:
+                    if estimator.target != initial_target:
+                        initial_target = estimator.target
+                        print("\n Predicting target {}".format(initial_target))
+                    target_distribution.append(estimator.target)
+                    print(estimator, "--->", estimator.target)
+                print("\nThe targets of the subgroups used in the prediction have the following distribution:")
+                pd.Series(target_distribution).value_counts().plot.pie()
+
+            elif is_regressor(self):
+                target_distribution = []
+                for estimator in covered_estimators:
+                    target_distribution.append(estimator.target)
+                    print(estimator, "--->", estimator.target)
+                print("\nThe targets of the subgroups used in the prediction have the following distribution:")
+                pd.Series(target_distribution).hist(bins=self.balance_bins)
+            else:
+                msg = "Can only show decision for Classifier or Regression models"
+                raise ValueError(msg)
         else:
             print("No subgroups cover this example. The default prediction is used.")
 
@@ -425,46 +440,40 @@ class RandomSubgroupClassifier(SubgroupPredictorBase, ClassifierMixin):
 
     def __init__(self,
                  n_estimators=100,
-                 max_depth=1,
+                 max_depth=2,
                  max_features="auto",
                  bootstrap=True,
-                 # oob_score=False,
                  n_jobs=None,
-                 # random_state=None,
-                 verbose=0,
-                 # class_weight=None,
+                 verbose=1,
                  max_samples=None,
+                 quality_function='standard',
                  quality_function_weight=0.5,
-                 search_strategy='bestfirst',
-                 result_set_size=1,
-                 intervals_only=True,
-                 n_bins=5):
-        super().__init__(
-            n_estimators=n_estimators,
-            # estimator_params=estimator_params,
-            bootstrap=bootstrap,
-            # oob_score=oob_score,
-            max_features=max_features,
-            n_jobs=n_jobs,
-            # random_state=random_state,
-            verbose=verbose,
-            # class_weight=class_weight,
-            max_samples=max_samples,
-            search_strategy=search_strategy,
-            result_set_size=result_set_size,
-            intervals_only=intervals_only)
+                 search_strategy='static',
+                 top_n=1,
+                 result_set_size=5,
+                 intervals_only=False,
+                 n_bins=5,
+                 binning='ef'):
+        super().__init__()
 
         self.n_estimators = n_estimators
         self.max_samples = max_samples
         self.max_depth = max_depth
+        self.max_features = max_features
+        self.bootstrap = bootstrap
         self.n_jobs = n_jobs
         self.verbose = verbose
+
+        # Subgroup Discovery parameters
+        self.quality_function = quality_function
         self.quality_function_weight = quality_function_weight
 
         self.search_strategy = search_strategy
+        self.top_n = top_n
         self.result_set_size = result_set_size
         self.intervals_only = intervals_only
         self.n_bins = n_bins
+        self.binning = binning
 
         self.classes_ = []
 
@@ -472,15 +481,32 @@ class RandomSubgroupClassifier(SubgroupPredictorBase, ClassifierMixin):
 
     def _subgroup_discovery(self, xy):
 
-        # self.target_id = random_state.choice(self.n_classes_)
-        self.target_id = np.random.choice(self.n_classes_)
-        xy.loc[:, 'target'] = (xy.loc[:, 'target'] == self.target_id)
+        # target_id = random_state.choice(self.n_classes_)
+        target_id = list(np.random.choice(self.n_classes_, 1, replace=False))[0]
+        # xy = xy.query("target == @target_id")
 
-        _target = ps.BinaryTarget(target_attribute='target', target_value=True)
+        _target = ps.BinaryTarget(target_attribute='target', target_value=target_id)
 
-        _search_space = ps.create_selectors(xy, ignore=['target'],
-                                            intervals_only=self.intervals_only,
-                                            nbins=self.n_bins)
+        if self.search_strategy == 'dynamic':
+            _search_space = []
+        elif self.search_strategy == 'static':
+            _search_space = get_search_space(xy, ignore=['target'],
+                                             intervals_only=self.intervals_only,
+                                             n_bins=self.n_bins)
+        else:
+            _search_space = ps.create_selectors(xy, ignore=['target'],
+                                                intervals_only=self.intervals_only,
+                                                nbins=self.n_bins)
+
+        if self.quality_function == 'standard':
+            _qf = ps.StandardQF(a=self.quality_function_weight)
+        elif self.quality_function == 'chisquared':
+            _qf = ps.ChiSquaredQF()
+        elif self.quality_function == 'ga':
+            _qf = ps.GeneralizationAware_StandardQF(a=self.quality_function_weight)
+        else:
+            msg = "Unknown nominal quality measure! Available options are: ['standard']"
+            raise ValueError(msg)
 
         task = ps.SubgroupDiscoveryTask(
             data=xy,
@@ -488,18 +514,20 @@ class RandomSubgroupClassifier(SubgroupPredictorBase, ClassifierMixin):
             search_space=_search_space,
             result_set_size=self.result_set_size,
             depth=self.max_depth,
-            qf=ps.StandardQF(a=self.quality_function_weight),
-            # qf=ps.ChiSquaredQF(),
+            qf=_qf,
             min_quality=0
         )
         result = self.ps_algorithm.execute(task)
 
         subgroup = self._check_estimator_output(result)
-        decoded_subgroup = decode_subgroup(subgroup)
+        if subgroup is not None:
+            subgroup_predictor_dict = [SubgroupPredictor(sg, target=target_id,
+                                                         alternative_target=None).to_dict() for sg in subgroup]
+            # decoded_subgroup_predictor = SubgroupPredictor(subgroup, target=int(target_id)).to_dict()
+        else:
+            subgroup_predictor_dict = None
 
-        target = int(self.target_id)
-
-        return decoded_subgroup, target
+        return subgroup_predictor_dict
 
     def predict(self, x):
         """
@@ -557,13 +585,33 @@ class RandomSubgroupClassifier(SubgroupPredictorBase, ClassifierMixin):
         n_samples = x.shape[0]
 
         proba = np.zeros((n_samples, self.n_classes_), dtype=np.float64)
-        for i in range(self.n_estimators):
-            proba[self.estimators_[i].covers(x), self.target[i]] += 1
+        for estimator in self.estimators_:
+            proba[estimator.covers(x), estimator.target] += 1
 
         proba = proba / self.n_estimators
 
         return proba
 
+    def feature_importances(self, absolute=False):
+
+        # Check is fit had been called
+        check_is_fitted(self)
+
+        counter = {target: {att: 0 for att in self.column_names} for target in range(self.n_classes_)}
+        for e in self.estimators_:
+            for f in e.get_features():
+                counter[e.target][f] += 1
+
+        counter_df = pd.DataFrame().from_dict(counter)
+        counter_df.columns = self.classes_
+
+        print(counter_df.sum(axis=1)/counter_df.values.sum())
+
+        if not absolute:
+            counter_df = counter_df / counter_df.sum(axis=0)
+
+        return counter_df
+
 
 class RandomSubgroupRegressor(SubgroupPredictorBase, RegressorMixin):
     """
@@ -651,89 +699,93 @@ class RandomSubgroupRegressor(SubgroupPredictorBase, RegressorMixin):
     >>> print(regr.predict([[0, 0, 0, 0]]))
     [-8.32987858]
     """
+
     def __init__(self,
                  n_estimators=100,
-                 # *,
-                 criterion="average",
-                 max_depth=1,
+                 max_depth=2,
                  max_features="auto",
                  bootstrap=True,
-                 # oob_score=False,
                  n_jobs=None,
-                 # random_state=None,
+                 random_state=None,
                  verbose=0,
-                 # class_weight=None,
                  max_samples=None,
                  quality_function_weight=0.5,
                  quality_function='absolute',
-                 search_strategy='bestfirst',
-                 result_set_size=1,
+                 criterion='average',
+                 search_strategy='static',
+                 top_n=1,
+                 result_set_size=5,
                  balance_bins=None,
-                 intervals_only=True,
-                 n_bins=5):
-        super().__init__(
-            n_estimators=n_estimators,
-            # estimator_params=estimator_params,
-            max_depth=max_depth,
-            bootstrap=bootstrap,
-            # oob_score=oob_score,
-            max_features=max_features,
-            n_jobs=n_jobs,
-            # random_state=random_state,
-            verbose=verbose,
-            # warm_start=warm_start,
-            # class_weight=class_weight,
-            max_samples=max_samples,
-            search_strategy=search_strategy,
-            result_set_size=result_set_size,
-            intervals_only=intervals_only)
+                 intervals_only=False,
+                 min_support=2,
+                 n_bins=5,
+                 binning='ef'):
+        super().__init__()
 
+        self.n_estimators = n_estimators
         self.max_depth = max_depth
+        self.max_features = max_features
+        self.bootstrap = bootstrap
+        self.n_jobs = n_jobs
+        self.random_state = random_state
+        self.verbose = verbose
+        self.max_samples = max_samples
 
+        # Subgroup Discovery parameters
         self.quality_function_weight = quality_function_weight
         self.quality_function = quality_function
         self.criterion = criterion
-
+        self.min_support = min_support
         self.search_strategy = search_strategy
+        self.top_n = top_n
         self.result_set_size = result_set_size
         self.intervals_only = intervals_only
         self.n_bins = n_bins
+        self.binning = binning
 
+        self.default_prediction = None
         self._estimator_type = "regressor"
 
+        # Experimental
         self.balance_bins = balance_bins
+        # np.random.seed(self.random_state)
+        # print(np.random.get_state()[1][0])
 
     def _subgroup_discovery(self, xy):
 
         if self.balance_bins:
-            discrete_target = pd.cut(xy.target, self.balance_bins)
-            # print(discrete_target)
-            # q = (1 - pd.Series(discrete_target).value_counts(normalize=True))
-            # q = q / np.sum(q)
-            # q2 = np.zeros(self.bins)
-            # for i in range(self.bins):
-            #     q2[discrete_target ==
-            unique_discrete_target = discrete_target.unique()
-            target_a = np.random.choice(unique_discrete_target)
-            target_b = np.random.choice(unique_discrete_target)
-            idx = (discrete_target == target_a) | (discrete_target == target_b)
-            xy = xy[idx]
-            # print(np.mean(xy.target))
+            # discrete_target = pd.qcut(xy.target, self.balance_bins, labels=False, duplicates='drop')
+            discrete_target = pd.cut(xy.target, self.balance_bins, labels=False)
+            (target_a, target_b) = np.random.choice(discrete_target.max()+1, 2, replace=False)
+            xy = xy[(discrete_target == target_a) | (discrete_target == target_b)]
+
+            # (idx1,) = np.where(discrete_target == target_a)
+            # (idx2,) = np.where(discrete_target == target_b)
+            # new_size = np.max([len(idx1), len(idx2)])
+            # idx = np.concatenate([np.random.choice(idx1, new_size), np.random.choice(idx2, new_size)])
+            # xy = xy.iloc[idx, :]
 
         _target = ps.NumericTarget('target')
 
-        _search_space = ps.create_selectors(xy, ignore=['target'],
-                                            intervals_only=self.intervals_only,
-                                            nbins=self.n_bins)
+        if self.search_strategy == 'dynamic':
+            _search_space = []
+        elif self.search_strategy == 'static':
+            _search_space = get_search_space(xy, ignore=['target'],
+                                             intervals_only=self.intervals_only,
+                                             n_bins=self.n_bins)
+        else:
+            _search_space = ps.create_selectors(xy, ignore=['target'],
+                                                intervals_only=self.intervals_only,
+                                                nbins=self.n_bins)
 
-        if self.quality_function == 'standard':
-            _qf = ps.StandardQFNumeric(a=self.quality_function_weight, estimator=self.criterion)
-        elif self.quality_function == 'old':
-            _qf = ps.OldStandardQFNumeric(a=self.quality_function_weight, estimator=self.criterion)
+        if self.quality_function == 'ps.standard':
+            _qf = ps.StandardQFNumeric(a=self.quality_function_weight)
+        elif self.quality_function == 'standard':
+            _qf = StandardNumeric(a=self.quality_function_weight, min_support=self.min_support)
         elif self.quality_function == 'absolute':
-            _qf = ps.AbsoluteQFNumeric(a=self.quality_function_weight, estimator=self.criterion)
+            _qf = AbsoluteNumeric(a=self.quality_function_weight, min_support=self.min_support)
         elif self.quality_function == 'kl':
-            _qf = ps.KLDivergenceNumeric(a=self.quality_function_weight)
+            _qf = KLDivergenceNumeric(a=self.quality_function_weight, min_support=self.min_support)
         else:
             msg = "Unknown numeric quality measure! Available options are: ['standard', 'absolute', 'kl']"
             raise ValueError(msg)
@@ -750,14 +802,33 @@ class RandomSubgroupRegressor(SubgroupPredictorBase, RegressorMixin):
         result = self.ps_algorithm.execute(task)
 
         subgroup = self._check_estimator_output(result)
-        decoded_subgroup = decode_subgroup(subgroup)
 
-        idx = subgroup[1].covers(xy)
-        target = np.array([np.mean(xy.target[idx]), np.mean(xy.target[np.invert(idx)])])
+        if subgroup is not None:
+            subgroup_predictor_dict = []
+            for sg in subgroup:
+                idx = sg[1].covers(xy)
+                if self.criterion == "average":
+                    target = float(np.mean(xy.target[idx]))
+                    alternative_target = float(np.mean(xy.target[np.invert(idx)]))
+                elif self.criterion == "median":
+                    target = float(np.median(xy.target[idx]))
+                    alternative_target = float(np.median(xy.target[np.invert(idx)]))
+                elif self.criterion == "maxmin":
+                    dataset_mean = np.mean(xy.target)
+                    if np.mean(xy.target[idx]) > dataset_mean:
+                        target = float(np.max(xy.target[idx]))
+                        alternative_target = float(np.max(xy.target[np.invert(idx)]))
+                    else:
+                        target = float(np.min(xy.target[idx]))
+                        alternative_target = float(np.min(xy.target[np.invert(idx)]))
+                subgroup_predictor_dict.append(SubgroupPredictor(sg, target=target,
+                                                                 alternative_target=alternative_target).to_dict())
+        else:
+            subgroup_predictor_dict = None
 
-        return decoded_subgroup, target
+        return subgroup_predictor_dict
 
-    def predict(self, x, all_contribute=False):
+    def predict(self, x):
 
         # class_type = self.classes_[0].dtype
         if not isinstance(x, pd.DataFrame):
@@ -775,28 +846,50 @@ class RandomSubgroupRegressor(SubgroupPredictorBase, RegressorMixin):
         predictions = np.zeros(n_samples)
         counter = np.zeros(n_samples)
 
-        for i in range(0, self.n_estimators):
-            idx = self.estimators_[i].covers(x)
-            predictions[idx] += self.target[i][0]
-
-            if all_contribute:
-                predictions[np.invert(idx)] += self.target[i][1]
-                counter += 1
-            else:
+        for estimator in self.estimators_:
+            idx = estimator.covers(x)
+            if any(idx):
                 counter[idx] += 1
+                for i, p in enumerate(estimator.predict(x)):
+                    if p is not None:
+                        predictions[i] += p
 
         predictions_missing = (counter == 0)
         if any(predictions_missing):
-            default_prediction = np.mean([e[1] for e in self.target])
-            predictions[predictions_missing] = default_prediction
+            # default_prediction = np.mean([e.alternative_target for e in self.estimators_])
+            # default_prediction = self.mean_target
+            predictions[predictions_missing] = self.default_prediction
             counter[predictions_missing] += 1
-            default_prediction_percent = round(np.mean(predictions_missing)*100, 0)
+            default_prediction_percent = round(np.mean(predictions_missing) * 100, 0)
             if default_prediction_percent > 5:
-                print("There are", str(default_prediction_percent), "% of default predictions. \n"
-                      "Consider increasing the number 'n_estimators'.")
+                if self.verbose:
+                    msg = '''There were {}% default predictions.
+                             Consider increasing 'n_estimators' or using 
+                             'balance_bins' > 2 '''.format(default_prediction_percent)
+                    # warnings.warn(msg)
+                    print(msg)
 
         predictions = predictions / counter
 
         # predictions.astype(class_type)
 
         return predictions
+
+    def feature_importances(self, absolute=False):
+
+        # Check is fit had been called
+        check_is_fitted(self)
+
+        counter = {att: 0 for att in self.column_names}
+        for e in self.estimators_:
+            for f in e.get_features():
+                counter[f] += 1
+
+        counter_df = pd.Series(counter)
+
+        print(counter_df/len(self.estimators_))
+
+        if not absolute:
+            counter_df = counter_df / counter_df.sum(axis=0)
+
+        return counter_df
\ No newline at end of file
